"""
Codewriters for Avro logical types.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from typing import Optional, List

from pyavro_gen.codewriters.core import ClassWriter, WritingMode, Dependency


class DatetimeClassWriter(ClassWriter):
    """
    Codewriter for datetime/timestamps.
    Can deal with Avro's millis and micros.
    """

    def __init__(self, timespec: Optional[str] = None):
        self.dep = ClassWriter('datetime.datetime')
        self.timespec = timespec

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            if self.timespec:
                return 'datetime.fromisoformat(datetime.now(tz=utc)' \
                       '.isoformat(timespec="milliseconds"))'
            return 'datetime.now(tz=utc)'
        return 'datetime'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return ClassWriter('pytz.utc').as_dependency(mode) + \
                   self.dep.as_dependency(mode)
        return self.dep.as_dependency(mode)


class DateClassWriter(ClassWriter):
    """
    Codewriter for dates.
    """

    def __init__(self) -> None:
        self.dep = ClassWriter('datetime.date')
        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            return 'datetime.now(tz=utc).date()'
        return 'date'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return ClassWriter('pytz.utc').as_dependency(mode) + \
                   self.dep.as_dependency(mode) + \
                   ClassWriter('datetime.datetime').as_dependency(mode)
        return self.dep.as_dependency(mode)


class UUIDClassWriter(ClassWriter):
    """
    Codewriter for UUIDs
    """

    def __init__(self) -> None:
        self.dep = ClassWriter('uuid.UUID')

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            return 'uuid4()'
        return 'UUID'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return ClassWriter('uuid.uuid4').as_dependency(mode)
        return self.dep.as_dependency(mode)



"""
Module that define the core classes for codewriters performing python classes generation
from Avro schemas.
"""

import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass, field as dataclasses_field
from enum import Enum
from typing import List, Optional, Set, Callable, cast

from isort import code

from pyavro_gen.codewriters.utils import class_name, namespace_name, j, ind, to_snake_case

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class WritingMode(Enum):
    """
    Writing mode for all codewriters
    """
    GENERATION = 'GENERATION'
    GENERATION_INDEPENDENT = 'GENERATION_INDEPENDENT'
    TYPE = 'TYPE'
    TEST = 'TEST'
    TEST_ATTRIBUTE = 'TEST_ATTRIBUTE'
    PARAMETER = 'PARAMETER'
    DOCSTRING_PARAM = 'DOCSTRING_PARAM'
    DOCSTRING_RAISES = 'DOCSTRING_RAISES'
    INIT = 'INIT'


@dataclass(frozen=True)
class Dependency:  # pylint: disable=R0903
    """
    A codewriter to express dependencies
    """
    where: str
    what: str
    prefix: Optional[str]

    def write(self, mode: WritingMode = WritingMode.GENERATION) -> str:
        """
        Returns this dependency as str, using the specified writing mode

        :param mode: The writing mode
        :return: The str form of the dependency
        """
        if mode == WritingMode.INIT:
            return \
                f'from {self.prefix}.{self.where}.{to_snake_case(self.what)} import {self.what}\n'
        if self.prefix:
            return f'from {self.prefix}.{self.where} import {self.what}\n'

        return f'from {self.where} import {self.what}\n'


class Writer(ABC):
    """
    Base class for all codewriters.
    """
    output_prefix: str = ''

    @abstractmethod
    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        """
        Returns this entity as str, using the specified writing mode

        :param mode: The writing mode
        :param with_indentation: How much to indent
        :return: The str form of the entity
        """

    # pylint: disable=R0201,W0613
    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        """
        Return this entity as a list of dependencies, using the specified writing mode.

        :param mode: The writing mode
        :return: A list of dependencies
        """
        return []


@dataclass
class Comment(Writer):
    """
    A codewriter for code comments.
    """
    value: Optional[str] = None

    def _optional_write(self, func: Callable[[str], str]) -> str:
        if self.value:
            return func(self.value)

        return ''

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        return self._optional_write(lambda s: '  # ' + self.value)  # type: ignore


@dataclass
class Docstring(Comment):
    """
    A codewriter for docstrings.
    """
    value: Optional[str] = None

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TYPE:
            return self._optional_write(lambda s: j('\n', f'#: {self.value}', '\n'))

        return self._optional_write(lambda s:
                                    j('\n',
                                      '"""', '\n',
                                      self.value, '\n',
                                      '"""', '\n'
                                      )
                                    )


# pylint: disable=R0902
class ClassWriter(Writer):
    """
    The main codewriter. It is an abstraction for an entity that writes Python classes.
    """

    def __init__(self,
                 fully_qualified_name: str,
                 doc: Optional[str] = None,
                 prefix: Optional[str] = None) -> None:

        self.name: str = class_name(fully_qualified_name)
        self.namespace: str = namespace_name(fully_qualified_name)
        self.doc: Docstring = Docstring(doc)
        self.prefix: Optional[str] = prefix
        self.comment: Comment = Comment()

        self.external_dependencies: List[str] = []
        self.decorators: List[Decorator] = []
        self.extensions: List[Extension] = []
        self.attributes: List[Attribute] = []
        self.methods: List[Method] = []

    def all_dependencies(self, mode: WritingMode) -> Set[Dependency]:
        """
        Return the set of dependencies this class depends on.

        :param mode: The writing mode
        """
        if mode == WritingMode.TEST:
            return {dep for a in self.attributes for dep in a.as_dependency(mode)}
        return set(
            # [d for dep in self.external_dependencies
            #  for d in ClassWriter(dep, prefix=self.prefix).as_dependency(mode)]
            [dep for d in self.decorators for dep in d.as_dependency(mode)]
            + [dep for e in self.extensions for dep in e.as_dependency(mode)]
            + [dep for a in self.attributes for dep in a.as_dependency(mode)]
            + [dep for m in self.methods for dep in m.as_dependency(mode)]
        )

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST and self.prefix:
            return [Dependency(self.namespace, self.name + 'Factory', self.prefix + '_test')]

        return [Dependency(self.namespace, self.name, self.prefix)]

    # pylint: disable=R0911
    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:

        if mode == WritingMode.TYPE:
            return self.name

        if mode == WritingMode.DOCSTRING_RAISES:
            return f':raises {".".join((self.namespace, self.name))}'

        if mode == WritingMode.GENERATION:
            if not self.attributes:
                if not self.methods:
                    attributes = j("\n", ind(1), "pass")
                else:
                    attributes = ''
            else:
                attributes = '\n'.join(a.write(mode) for a in self.attributes)

            attr_meth_sep = '\n' if self.methods and self.attributes else ''

            methods = '\n'.join(m.write(mode) for m in self.methods) if self.methods else ''

            extensions = f"({', '.join(e.write(mode) for e in self.extensions)})" \
                if self.extensions else ''

            decorators = '\n'.join(dec.write(mode) for dec in self.decorators) + '\n' \
                if self.decorators else ''

            return textwrap.indent(
                j(
                    '\n\n',
                    decorators,
                    f'class {self.name}{extensions}:',
                    textwrap.indent(
                        j(
                            self.doc.write(mode),
                            attributes,
                            attr_meth_sep,
                            methods,
                            '\n'
                        ),
                        ind(1)
                    )
                ),
                ind(with_indentation)
            )

        if mode == WritingMode.GENERATION_INDEPENDENT:
            all_deps = {dep.write() for dep in self.all_dependencies(mode)}

            content = j(
                j(*all_deps),
                self.write(WritingMode.GENERATION, with_indentation)
            )

            return code(content)

        if mode == WritingMode.TEST_ATTRIBUTE:
            return f'{self.name}Factory()'

        if mode == WritingMode.TEST:
            all_deps = {
                dep.write() for dep in self.all_dependencies(mode)
            }.union({
                # Factory boy dependencies
                Dependency('factory', 'Factory', None).write(),
                Dependency('factory', 'lazy_attribute', None).write(),
            }).union({
                # Itself in its standard (non-factory) form
                dep.write(WritingMode.TEST)
                for dep in self.as_dependency(WritingMode.GENERATION)
            })

            class_content = textwrap.indent(
                j(
                    '\n\n',
                    f'class {self.name}Factory(Factory):',
                    textwrap.indent(
                        j(
                            f'\nclass Meta:\n{ ind(1)}model = {self.name}\n',
                            self.get_testing_attributes(),
                            '\n'
                        ),
                        ind(1)
                    )
                ),
                ind(with_indentation)
            )

            content = j(
                j(*all_deps),
                class_content
            )

            return code(content)

        return ''

    def get_testing_attributes(self) -> str:
        """
        Returns the string version of attributes of this class for writing mode == TEST.
        NOTE: Enums redefine it.
        """
        if not self.attributes:
            return j("\n", ind(1), "pass")

        return '\n'.join(
            a.write(WritingMode.TEST) for a in self.attributes
            if not a.name.startswith('_')  # filtering schema class variable
        )


@dataclass
class Decorator(Writer):
    """
    Codewriter for decorators.
    """
    value: str
    dependency: Optional[ClassWriter] = None

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.GENERATION:
            return self.value
        return ''

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if self.dependency:
            return self.dependency.as_dependency(mode)
        return []


class Extension(Writer):
    """
    Codewriter for classes extending other classes.
    """

    def __init__(self, fully_qualified_name: str) -> None:
        self.dependency: ClassWriter = ClassWriter(fully_qualified_name)

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        return self.dependency.write(WritingMode.TYPE)

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        return self.dependency.as_dependency(mode)


@dataclass
class Attribute(Writer):
    """
    Codewriter for an attribute of a class.
    An attribute may have a type, which is expressed by a ClassWriter.
    If it does *not* have a type, it's one of Python standard attributes (e.g. 'self', 'cls', etc.).
    """
    name: str
    type: Optional[ClassWriter] = None
    default_value: Optional[str] = None
    doc: Docstring = dataclasses_field(default_factory=lambda: Docstring())

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        default = f' = {self.default_value}' if self.default_value else ''

        if not self.type:
            if mode == WritingMode.PARAMETER:
                return self.name
            return f'{self.name}{default}'

        if mode == WritingMode.TEST:
            test_attribute = self.type.write(WritingMode.TEST_ATTRIBUTE)
            return f'{self.name} = lazy_attribute(lambda x: {test_attribute})'

        if mode == WritingMode.PARAMETER:
            return f'{self.name}: {self.type.write(WritingMode.TYPE)}{default}'

        if mode == WritingMode.DOCSTRING_PARAM:
            return f':param {self.name}: {self.doc.value}'

        if isinstance(self.type, EnumValueWriter):
            name = self.name
            if self.name.startswith('_'):
                # enums names cannot start with '_'
                while name.startswith('_'):
                    name = name[1:]
            str_attribute = f'{name} = {self.type.write(WritingMode.TYPE)}{default}'
        else:
            str_attribute = f'{self.name}: {self.type.write(WritingMode.TYPE)}{default}'

        return textwrap.indent(
            j(
                self.doc.write(WritingMode.TYPE, with_indentation),
                str_attribute,
                self.type.comment.write(WritingMode.TYPE)
            ),
            ind(with_indentation)
        )

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if not self.type:
            return []
        return self.type.as_dependency(mode)


@dataclass
class Method(Writer):
    """
    Codewriter for a method of a class.
    """

    name: str
    parameters: List[Attribute]
    return_type: ClassWriter
    doc: Docstring = dataclasses_field(default_factory=lambda: Docstring())
    errors: List[ClassWriter] = dataclasses_field(default_factory=list)
    body: Optional[str] = None
    additional_dependencies: List[ClassWriter] = dataclasses_field(default_factory=list)
    decorators: List[Decorator] = dataclasses_field(default_factory=list)

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.GENERATION:
            str_body = self.body if self.body else 'pass'

            if len(self.parameters) <= 1:
                sep_params = ''
                par_indent = 0
            else:
                sep_params = '\n'
                par_indent = 2

            str_doc = j(self.doc.value, sep_params * 2) if self.doc else ''
            sep_error = '\n' if self.errors else ''

            str_decorators = '\n'.join(
                dec.write(mode) for dec in self.decorators  # pylint: disable=E1133
            ) + '\n' \
                if self.decorators else ''

            str_doc_params = (f'{p.write(WritingMode.DOCSTRING_PARAM)}' for p in self.parameters)
            str_doc_params = (s for s in str_doc_params if s)

            return j(
                # spacer
                '\n',

                # decorators
                str_decorators,

                # begin definition
                j(f'def { to_snake_case(self.name)}(', sep_params),

                # arguments
                textwrap.indent(
                    ',\n'.join([p.write(WritingMode.PARAMETER) for p in self.parameters]),
                    ind(par_indent)
                ),

                # definition_end
                f'{sep_params}) -> {self.return_type.write(WritingMode.TYPE)}:\n',

                # doc begin
                textwrap.indent(j('"""\n', str_doc), ind(1)),

                # doc params
                textwrap.indent(
                    '\n'.join(str_doc_params),
                    ind(1)
                ),

                # doc raises
                textwrap.indent(
                    sep_error + '\n'.join(
                        (f'{p.write(WritingMode.DOCSTRING_RAISES)}'
                         for p in self.errors)),  # pylint: disable=E1133
                    ind(1)
                ),

                # doc end and body
                textwrap.indent(f'\n"""\n{str_body}', ind(1)),
            )

        return ''

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        return [dep for parameter in self.parameters for dep in parameter.as_dependency(mode)] + \
               self.return_type.as_dependency(mode) + \
               [dep for additional_dependency
                in self.additional_dependencies  # pylint: disable=E1133
                for dep in additional_dependency.as_dependency(mode)] + \
               [dep for decorator
                in self.decorators  # pylint: disable=E1133
                for dep in decorator.as_dependency(mode)]


class EnumClassWriter(ClassWriter):
    """
    Codewriter for enum classes.
    """

    def __init__(self,
                 fully_qualified_name: str,
                 doc: Optional[str] = None,
                 prefix: Optional[str] = None) -> None:
        super().__init__(fully_qualified_name, doc, prefix)

        self.extensions = [
            Extension('enum.Enum')
        ]

    def get_testing_attributes(self) -> str:
        return f'value = lazy_attribute(lambda x: fake.enum_with_schema({self.name}))'

    def all_dependencies(self, mode: WritingMode) -> Set[Dependency]:
        if mode == WritingMode.TEST:
            return {Dependency(self.output_prefix, 'fake', None)}

        return super().all_dependencies(mode).union({Dependency('enum', 'nonmember', None)})


class EnumValueWriter(ClassWriter):
    """
    Codewriter for enum values.
    """

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        return f"'{self.name}'"

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        return []


class SimpleTypeClassWriter(ClassWriter):
    """
    Codewriter for simple types such as int, str, bool, etc.
    """

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            if self.name == 'None':
                return 'None'
            return f'fake.py{self.name}()'
        return self.name

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return [Dependency(self.output_prefix, 'fake', None)]
        return []


class ClassVarClassWriter(ClassWriter):
    """
    Codewriter for class variables (e.g. the _schema property for Avro schemas (str))
    """

    def __init__(self,
                 writer: ClassWriter,
                 additional_dependencies: Optional[List[ClassWriter]] = None
                 ) -> None:
        self.class_writer: ClassWriter = writer
        self.dep = ClassWriter('typing.ClassVar')
        self.additional_dependencies: List[ClassWriter] = \
            additional_dependencies if additional_dependencies else []

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        return f'ClassVar[{self.class_writer.write(WritingMode.TYPE)}]'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return []
        return self.class_writer.as_dependency(mode) + self.dep.as_dependency(mode) + \
               [dep for additional_dependency
                in self.additional_dependencies  # pylint: disable=E1133
                for dep in additional_dependency.as_dependency(mode)]



"""
A module to recursively compare any object considering the types of all their constituents
and using two strategies to perform almost equality checks of floats.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

import unittest
from collections import namedtuple, OrderedDict
from dataclasses import dataclass
from typing import Any, cast, Dict


def are_almost_equal(  # pylint: disable=R0911,R0912
        obj1: Any,
        obj2: Any,
        max_abs_ratio_diff: float,
        max_abs_diff: float
) -> bool:
    """
    Compares two objects by recursively walking them trough. Equality is as usual except for floats.
    Floats are compared according to the two measures defined below.

    :param obj1: The first object.
    :param obj2: The second object.
    :param max_abs_ratio_diff: The maximum allowed absolute value of the difference.
    `abs(1 - (obj1 / obj2)` and vice-versa if obj2 == 0.0. Ignored if < 0.
    :param max_abs_diff: The maximum allowed absolute difference `abs(obj1 - obj2)`. Ignored if < 0.
    :return: Whether the two objects are almost equal.
    """
    if type(obj1) is not type(obj2):
        return False

    composite_type_passed = False

    if hasattr(obj1, '__slots__'):
        if len(obj1.__slots__) != len(obj2.__slots__):
            return False
        if any(not are_almost_equal(getattr(obj1, s1), getattr(obj2, s2),
                                    max_abs_ratio_diff, max_abs_diff)
               for s1, s2 in zip(sorted(obj1.__slots__), sorted(obj2.__slots__))):
            return False

        composite_type_passed = True

    if hasattr(obj1, '__dict__'):
        if len(obj1.__dict__) != len(obj2.__dict__):
            return False
        if any(not are_almost_equal(k1, k2, max_abs_ratio_diff, max_abs_diff)
               or not are_almost_equal(v1, v2, max_abs_ratio_diff, max_abs_diff)
               for ((k1, v1), (k2, v2))
               in zip(sorted(obj1.__dict__.items()), sorted(obj2.__dict__.items()))
               if not k1.startswith('__')):  # avoid infinite loops
            return False

        composite_type_passed = True

    if isinstance(obj1, dict):
        if len(obj1) != len(obj2):
            return False
        if any(not are_almost_equal(k1, k2, max_abs_ratio_diff, max_abs_diff)
               or not are_almost_equal(v1, v2, max_abs_ratio_diff, max_abs_diff)
               for ((k1, v1), (k2, v2)) in zip(sorted(obj1.items()), sorted(obj2.items()))):
            return False

    elif any(issubclass(obj1.__class__, c) for c in (list, tuple, set)):
        if len(obj1) != len(obj2):
            return False
        if any(not are_almost_equal(v1, v2, max_abs_ratio_diff, max_abs_diff)
               for v1, v2 in zip(obj1, obj2)):
            return False

    elif isinstance(obj1, float):
        if obj1 == obj2:
            return True

        if max_abs_ratio_diff > 0:  # if max_abs_ratio_diff < 0, max_abs_ratio_diff is ignored
            if obj2 != 0:
                if abs(1.0 - (obj1 / obj2)) > max_abs_ratio_diff:
                    return False
            else:  # if both == 0, we already returned True
                if abs(1.0 - (obj2 / obj1)) > max_abs_ratio_diff:
                    return False
        if 0 < max_abs_diff < abs(obj1 - obj2):  # if max_abs_diff < 0, max_abs_diff is ignored
            return False
        return True

    else:
        if not composite_type_passed:
            return cast(bool, obj1 == obj2)

    return True


class EqualityTest(unittest.TestCase):
    """
    Test cases for are_almost_equal
    """

    def test_floats(self) -> None:
        """
        Test float comparison strategies
        """
        obj1 = ('hi', 3, 3.4)
        obj2 = ('hi', 3, 3.400001)
        self.assertTrue(are_almost_equal(obj1, obj2, 0.0001, 0.0001))
        self.assertFalse(are_almost_equal(obj1, obj2, 0.00000001, 0.00000001))

    def test_ratio_only(self) -> None:
        """
        Tests ratio comparison strategy only
        """
        obj1 = ['hey', 10000, 123.12]
        obj2 = ['hey', 10000, 123.80]
        self.assertTrue(are_almost_equal(obj1, obj2, 0.01, -1))
        self.assertFalse(are_almost_equal(obj1, obj2, 0.001, -1))

    def test_diff_only(self) -> None:
        """
        Tests diff comparison strategy only
        """
        obj1 = ['hey', 10000, 1234567890.12]
        obj2 = ['hey', 10000, 1234567890.80]
        self.assertTrue(are_almost_equal(obj1, obj2, -1, 1))
        self.assertFalse(are_almost_equal(obj1, obj2, -1, 0.1))

    def test_both_ignored(self) -> None:
        """
        Tests that both strategies are ignored
        """
        obj1 = ['hey', 10000, 1234567890.12]
        obj2 = ['hey', 10000, 0.80]
        obj3 = ['hi', 10000, 0.80]
        self.assertTrue(are_almost_equal(obj1, obj2, -1, -1))
        self.assertFalse(are_almost_equal(obj1, obj3, -1, -1))

    def test_different_lengths(self) -> None:
        """
        Tests that returns false if containers have different lengths
        """
        obj1 = ['hey', 1234567890.12, 10000]
        obj2 = ['hey', 1234567890.80]
        self.assertFalse(are_almost_equal(obj1, obj2, 1, 1))

    def test_classes(self) -> None:
        """
        Test comparing function using class arguments.
        Monkey patched classes should also be supported.
        """

        # pylint: disable=C0111,C0103,R0903
        class A:
            d = 12.3

            def __init__(self, a: float, b: str, c: Dict[Any, Any]) -> None:
                self.attr_a = a
                self.attr_b = b
                self.attr_c = c

        obj1: A = A(2.34, 'str', {1: 'hey', 345.23: [123, 'hi', 890.12]})
        obj2: A = A(2.34, 'str', {1: 'hey', 345.231: [123, 'hi', 890.121]})
        self.assertTrue(are_almost_equal(obj1, obj2, 0.1, 0.1))
        self.assertFalse(are_almost_equal(obj1, obj2, 0.0001, 0.0001))

        # This should fail even if no float inequality is considered because we add a new attribute
        # pylint: disable=W0201
        obj2.hello = 'hello'  # type: ignore
        self.assertFalse(are_almost_equal(obj1, obj2, -1, -1))

    def test_namedtuples(self) -> None:
        """
        Test support for namedtuples
        """
        B = namedtuple('B', ['x', 'y'])
        obj1 = B(3.3, 4.4)
        obj2 = B(3.4, 4.5)
        self.assertTrue(are_almost_equal(obj1, obj2, 0.2, 0.2))
        self.assertFalse(are_almost_equal(obj1, obj2, 0.001, 0.001))

    def test_classes_with_slots(self) -> None:
        """
        Tests support for classes with no __dict__, that use __slots__
        """

        # pylint: disable=C0111,C0103,R0903
        class C:
            __slots__ = ['attr_a', 'attr_b']

            def __init__(self, a: float, b: float) -> None:
                self.attr_a = a
                self.attr_b = b

        obj1: C = C(3.3, 4.4)
        obj2: C = C(3.4, 4.5)
        self.assertTrue(are_almost_equal(obj1, obj2, 0.3, 0.3))
        self.assertFalse(are_almost_equal(obj1, obj2, -1, 0.01))

    def test_dataclasses(self) -> None:
        """
        Tests support for dataclasses
        """

        # pylint: disable=C0111,C0103
        @dataclass
        class D:
            s: str
            i: int
            f: float

        # pylint: disable=C0111,C0103
        @dataclass
        class E:
            f2: float
            f4: str
            d: D

        obj1 = E(12.3, 'hi', D('hello', 34, 20.01))
        obj2 = E(12.1, 'hi', D('hello', 34, 20.0))
        self.assertTrue(are_almost_equal(obj1, obj2, -1, 0.4))
        self.assertFalse(are_almost_equal(obj1, obj2, -1, 0.001))

        obj3 = E(12.1, 'hi', D('ciao', 34, 20.0))
        self.assertFalse(are_almost_equal(obj2, obj3, -1, -1))

    def test_ordereddict(self) -> None:
        """
        Tests support for OrderedDict
        """
        obj1 = OrderedDict({1: 'hey', 345.23: [123, 'hi', 890.12]})
        obj2 = OrderedDict({1: 'hey', 345.23: [123, 'hi', 890.0]})
        self.assertTrue(are_almost_equal(obj1, obj2, 0.01, -1))
        self.assertFalse(are_almost_equal(obj1, obj2, 0.0001, -1))



#!/usr/bin/env python

"""
Command line entrypoint for pyavro-gen
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

import argparse

from avro_preprocessor.avro_paths import AvroPaths
from pyavro_gen.generator import AvroGenerator

if __name__ == '__main__':
    PARSER = argparse.ArgumentParser(
        description='A typed class generator for Avro Schemata.')

    PARSER.add_argument('-i', '--input-path', dest='input_path', type=str, required=True)
    PARSER.add_argument('-o', '--output-path', dest='output_path', type=str, required=True)

    PARSER.add_argument('-b', '--base-namespace', dest='base_namespace', type=str, required=True)

    TYPES = PARSER.add_mutually_exclusive_group(required=True)
    TYPES.add_argument('-t', '--types-namespace', dest='types_namespace', type=str)
    TYPES.add_argument(
        '-n', '--no-types-namespace', dest='types_namespace', action='store_const', const=None)

    RPC = PARSER.add_mutually_exclusive_group(required=True)
    RPC.add_argument('-r', '--rpc-namespace', dest='rpc_namespace', type=str)
    RPC.add_argument(
        '-e', '--no-rpc-namespace', dest='rpc_namespace', action='store_const', const=None)

    PARSER.add_argument(
        '-ie', '--input-schema-file-extension', dest='input_schema_file_extension', type=str,
        default='avsc')

    PARSER.add_argument('-v', '--verbose', dest='verbose', action='store_true')
    PARSER.set_defaults(verbose=False)

    # INDENT = PARSER.add_mutually_exclusive_group()
    # INDENT.add_argument('-ci', '--classes_indent', dest='classes_indent', type=int, default=4)

    AVAILABLE_MODULES = ' '.join(AvroGenerator.available_preprocessing_modules.keys())
    PARSER.add_argument('-m', '--modules', dest='modules', nargs='*', default=None,
                        help='Available modules: {}'.format(AVAILABLE_MODULES))

    ARGS = PARSER.parse_args()

    AVRO_GENERATOR: AvroGenerator = AvroGenerator(
        AvroPaths(
            input_path=ARGS.input_path,
            output_path=ARGS.output_path,
            base_namespace=ARGS.base_namespace,
            types_namespace=ARGS.types_namespace,
            rpc_namespace=ARGS.rpc_namespace,
            input_schema_file_extension=ARGS.input_schema_file_extension,
        ),
        verbose=ARGS.verbose,
        # classes_indent=ARGS.classes_indent,
    )

    AVRO_GENERATOR.process(ARGS.modules)



"""
Codewriters for container classes.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from typing import List

from pyavro_gen.codewriters.core import ClassWriter, WritingMode, Dependency
from pyavro_gen.codewriters.utils import j


class UnionClassWriter(ClassWriter):
    """
    Codewriter for Union and Optionals.
    """

    def __init__(self, writers: List[ClassWriter]) -> None:
        if len(writers) == 2 and any(t.name == 'None' for t in writers):
            self.class_writers = [w for w in writers if w.name != 'None']
            self.type_prefix = 'Optional'
            self.dep = ClassWriter('typing.Optional')
        else:
            self.class_writers = writers
            self.type_prefix = 'Union'
            self.dep = ClassWriter('typing.Union')

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            ta_str = [w.write(WritingMode.TEST_ATTRIBUTE) for w in self.class_writers]
            rand_len = len(self.class_writers) - 1
            if self.type_prefix == 'Optional':
                ta_str.append('None')
                rand_len += 1
            test_attribute = ", ".join(ta_str)
            return f'[{test_attribute}][randint(0, {rand_len})]'

        union_arguments = ", ".join(w.write(WritingMode.TYPE) for w in self.class_writers)
        return f'{self.type_prefix}[{union_arguments}]'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        deps = [dep for writer in self.class_writers for dep in writer.as_dependency(mode)]
        if mode == WritingMode.TEST:
            return deps + ClassWriter('random.randint').as_dependency(mode)
        return deps + self.dep.as_dependency(mode)


class ListClassWriter(ClassWriter):
    """
    Codewriter for Lists.
    """

    def __init__(self, writer: ClassWriter) -> None:
        self.class_writer: ClassWriter = writer
        self.dep = ClassWriter('typing.List')

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            test_attribute = self.class_writer.write(WritingMode.TEST_ATTRIBUTE)
            return f'[{test_attribute} for _ in range(randint(1, 5))]'
        return f'List[{self.class_writer.write(WritingMode.TYPE)}]'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return self.class_writer.as_dependency(mode) + \
                   ClassWriter('random.randint').as_dependency(mode)
        return self.class_writer.as_dependency(mode) + self.dep.as_dependency(mode)


class MapClassWriter(ClassWriter):
    """
    Codewriter for Maps.
    """

    def __init__(self, writer: ClassWriter) -> None:
        self.class_writer: ClassWriter = writer
        self.dep = ClassWriter('typing.Dict')

        super().__init__(self.write(WritingMode.GENERATION))

    def write(self, mode: WritingMode = WritingMode.GENERATION, with_indentation: int = 0) -> str:
        if mode == WritingMode.TEST_ATTRIBUTE:
            test_attribute = self.class_writer.write(WritingMode.TEST_ATTRIBUTE)
            return j('{', f'fake.pystr(): {test_attribute} for _ in range(randint(3, 10))', '}')
        # All Avro Map keys are strings
        return f'Dict[str, {self.class_writer.write(WritingMode.TYPE)}]'

    def as_dependency(self, mode: WritingMode) -> List[Dependency]:
        if mode == WritingMode.TEST:
            return self.class_writer.as_dependency(mode) + \
                   ClassWriter('random.randint').as_dependency(mode)
        return self.class_writer.as_dependency(mode) + self.dep.as_dependency(mode)



"""
Utility functions to deal with string and class names.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

import re
from typing import Optional

indentation: int = 4


def namespace_name(fully_qualified_name: str) -> str:
    """
    Return the namespace name of a fully qualified class name.
    :param fully_qualified_name: The fully qualified name
    """
    tokens = fully_qualified_name.split('.')
    return '.'.join(tokens[:-1])


def class_name(fully_qualified_name: str) -> str:
    """
    Returns the class name of a fully qualified class name.

    :param fully_qualified_name: The fully qualified name
    """
    tokens = fully_qualified_name.split('.')
    return tokens[-1]


def to_snake_case(camel_str: str) -> str:
    """
    CamelCase str to snake_case

    :param camel_str: The str in input
    """
    capitalized_snake_str = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', camel_str)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', capitalized_snake_str).lower()


def to_camel_case(snake_str: str, except_first: bool = True) -> str:
    """
    snake_case string to CamelCase

    :param snake_str: The input str
    :param except_first: does not capitalise the first letter
    """
    components = snake_str.split('_')
    if except_first:
        # We capitalize the first letter of each component except the first one
        # with the 'title' method and join them together.
        return components[0] + ''.join(x.title() for x in components[1:])
    return ''.join(x.title() for x in components)


def ind(i: int = 1) -> str:
    """
    Returns a number of spaces corresponding to the indentation level in input

    :param i: The indentation level
    """
    return ' ' * indentation * i


def j(*strings: Optional[str]) -> str:
    """
    Quick helper to join strings

    :param strings: The strings to join
    """
    return ''.join(s for s in strings if s)



"""
A memory container for schemas and generated classes.
It also writes python classes to disk.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from collections import OrderedDict
from pathlib import Path
from typing import Dict, Set, List, Any  # pylint: disable=W0611

from avro_preprocessor.avro_paths import AvroPaths
from avro_preprocessor.schemas_container import SchemasContainer

from pyavro_gen.codewriters.core import WritingMode, Writer, ClassWriter  # pylint: disable=no-name-in-module
from pyavro_gen.codewriters.namespace import ClassItem, NamespaceWriter, get_testing_base, \
    get_testing_class_list
from pyavro_gen.codewriters.utils import namespace_name


# pylint: disable=R0914
class SchemaAndClassesContainer(SchemasContainer):  # type: ignore
    """
    A memory container for schemas and generated classes.
    It also writes python classes to disk.
    """

    def __init__(self, paths: AvroPaths, verbose: bool = True) -> None:
        super(SchemaAndClassesContainer, self).__init__(paths, verbose)

        self.output_writers: 'OrderedDict[str, ClassWriter]' = OrderedDict()

        self.output_prefix = Path(self.paths.output_path).name

    def _store_schema(self, schema: 'OrderedDict[Any, Any]', namespace: str) -> None:
        self.original[namespace] = schema

    def write_classes(self, mode: WritingMode = WritingMode.GENERATION_INDEPENDENT) -> None:
        """
        Writes python classes to disk.
        """

        namespace_writers: Dict[str, NamespaceWriter] = {}
        for name, class_writer in self.output_writers.items():
            namespace = namespace_name(name)
            if namespace not in namespace_writers:
                namespace_writers[namespace] = NamespaceWriter(namespace)
            namespace_writers[namespace].class_writers.append(class_writer)

        all_modes = [mode, WritingMode.TEST]

        all_testing_classes: List[ClassItem] = []

        for current_mode in all_modes:

            if current_mode == WritingMode.TEST:
                path = Path(self.paths.output_path)
                test_name = str(path.name) + '_test'
                output_path = path.parent.joinpath(Path(test_name))
            else:
                output_path = Path(self.paths.output_path)

            print('############### OUTPUT PATH:', output_path)

            Writer.output_prefix = output_path.name.replace('/', '')

            # let's cleanup the output schema directory first
            AvroPaths.reset_directory(output_path)

            if current_mode != WritingMode.TEST:
                main_init = output_path.joinpath('__init__.py')
                main_init.parent.mkdir(parents=True, exist_ok=True)

            parent_paths: Set[Path] = set()

            for namespace_writer in namespace_writers.values():
                print(namespace_writer.namespace, [w.name for w in namespace_writer.class_writers])

                fname = namespace_writer.namespace.replace('.', '/')
                init_file_path = output_path.joinpath(Path(fname))
                init_file_path.mkdir(parents=True, exist_ok=True)
                parent_paths = parent_paths.union(set(par for par in init_file_path.parents
                                                      if str(output_path) in str(par)))

                contents = namespace_writer.write(current_mode)
                self._write_to_disk(
                    {init_file_path.joinpath(fname): content for fname, content in contents.items()}
                )

                if current_mode == WritingMode.TEST:
                    all_testing_classes += namespace_writer.get_all_testing_classes()

            if current_mode == WritingMode.TEST:
                testing_base_content = get_testing_base()
                testing_base_path = output_path.joinpath(Path('__init__.py'))

                testing_class_list_content = get_testing_class_list(all_testing_classes)
                testing_class_list_path = output_path.joinpath(Path('testing_classes.py'))

                self._write_to_disk({
                    testing_base_path: testing_base_content,
                    testing_class_list_path: testing_class_list_content
                })

            for to_exclude in ['.', '..']:
                try:
                    parent_paths.remove(Path(to_exclude))
                except KeyError:
                    pass

            for path in sorted(parent_paths):
                self._init_python_module(Path(path))

    @staticmethod
    def _init_python_module(path: Path) -> None:
        path = path.joinpath('__init__.py')
        if not path.exists():
            path.write_text('')

    @staticmethod
    def _write_to_disk(contents: Dict[Path, str]) -> None:
        for fname, content in contents.items():
            fname.write_text(content)

    write_schemas = write_classes



"""
A generator for Python classes from Avro schemas.
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from collections import OrderedDict
from typing import List

from avro_preprocessor.preprocessor import AvroPreprocessor
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

from pyavro_gen.codewriters.core import WritingMode
from pyavro_gen.schema_and_classes_container import SchemaAndClassesContainer
from pyavro_gen.modules.avsc_schema_dependency_checker import AvscSchemaDependenciesChecker
from pyavro_gen.modules.fields_collector import FieldsCollector


class AvroGenerator(AvroPreprocessor):  # type: ignore
    """
    A generator for Python classes from Avro schemas.
    """
    preprocessing_modules = [
        AvscSchemaDependenciesChecker,
        FieldsCollector
    ]

    #: OrderedDict of module_name -> module_class
    available_preprocessing_modules = OrderedDict((
        (m.__name__, m) for m in preprocessing_modules
    ))

    def _get_schemas_container(self) -> SchemasContainer:
        return SchemaAndClassesContainer(self.paths, self.verbose)

    def process_with_modules(
            self,
            requested_modules: List[PreprocessorModule],
            mode: WritingMode = WritingMode.GENERATION_INDEPENDENT
    ) -> None:
        """
        Process all the schemas using the modules pipeline.
        files in the output path.

        :param requested_modules: Instances of the modules to use.
        :param mode: WritingMode.GENERATION_INDEPENDENT to create independent class files,
        WritingMode.GENERATION to put all content in `__init__.py` files.
        """

        schemas: SchemaAndClassesContainer = self.schemas

        schemas.read_schemas()
        modules = OrderedDict(
            (module.__class__.__name__, module) for module in requested_modules
        )
        self._do_process(modules)
        schemas.write_schemas(mode)



"""
A module to find dependencies among all records in all schemas.
"""
__author__ = "Nicola Bova"
__copyright__ = "Copyright 2019, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from collections import OrderedDict
from typing import cast

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.modules.schema_dependencies_checker import SchemaDependenciesChecker

from pyavro_gen.codewriters.utils import namespace_name


class AvscSchemaDependenciesChecker(SchemaDependenciesChecker):  # type: ignore
    """
    Stores schema dependencies - exception thrown if it detects cycles.

    NOTE: Also all Records defined within schemas are considered.
    """

    def store_dependencies_of_field(self, node: Avro.Node) -> None:
        """
        Store external_dependencies of other records in a node in a private dict
        :param node: The input node
        """

        if isinstance(node, str):
            if self.ancestors and '.' in node:

                # it has to be a 'name' field or part of a list
                anc = self.ancestors[-1].key
                if anc == Avro.Name or isinstance(anc, int):

                    dependent_ancestor = self._find_ancestor()
                    if dependent_ancestor:
                        self.record_dependencies_graph.add_edge(dependent_ancestor, node)

        if isinstance(node, OrderedDict):
            if Avro.Name in node:
                if Avro.Namespace in node:
                    dep = node[Avro.Namespace] + '.' + node[Avro.Name]
                elif '.' in node[Avro.Name]:
                    dep = node[Avro.Name]
                elif Avro.Fields in node or Avro.Symbols in node:
                    dep = namespace_name(self.current_schema_name) + '.' + node[Avro.Name]
                else:
                    return

                dependent_ancestor = self._find_ancestor()
                self.record_dependencies_graph.add_edge(dependent_ancestor, dep)

    def _find_ancestor(self) -> str:
        dependent_ancestor = None
        for ancestor in reversed(self.ancestors):

            if Avro.Name in ancestor.node \
                    and ancestor.node[Avro.Type] == Avro.Record \
                    and isinstance(ancestor.node, OrderedDict):

                if Avro.Namespace in ancestor.node:
                    dependent_ancestor = \
                        ancestor.node[Avro.Namespace] + '.' + ancestor.node[Avro.Name]
                elif '.' in ancestor.node[Avro.Name]:
                    dependent_ancestor = ancestor.node[Avro.Name]
                else:
                    dependent_ancestor = \
                        namespace_name(self.current_schema_name) \
                        + '.' + ancestor.node[Avro.Name]
                break

        if not dependent_ancestor:
            return cast(str, self.current_schema_fully_qualified_name)

        return cast(str, dependent_ancestor)
